/*
 * basic_map.h
 *
 * Provides a simple occupancy map with occupied, free cells.
 * To aid in converting from the probability grid, a functors are provided.
 *
 *  Created on: 16/08/2011
 *      Author: daver
 */

#ifndef BASIC_MAP_H_
#define BASIC_MAP_H_

#include <tr1/functional>
#include <boost/functional/hash.hpp>
#include <boost/multi_array.hpp>
#include "bushfire.h"

// Assume occupied if the cell is unknown and has a path of unknown
// adjacents to a border.
struct BasicMapCell
{
  typedef enum { free=0, occupied=1 } t_occupancy;
  t_occupancy occupancy_;


  inline bool isFree() const { return (occupancy_ == free); }
  inline bool isOccupied() const { return (occupancy_ == occupied); }

  BasicMapCell(): occupancy_(free) {}
  BasicMapCell(t_occupancy occ): occupancy_(occ) {}

  inline bool operator==(const BasicMapCell& o) { return (occupancy_ == o.occupancy_); }
  inline bool operator!=(const BasicMapCell& o) { return (occupancy_ != o.occupancy_); }
};


// Conversion functor to convert probability points on the occupancy grid
// to a occupied/free based on a threshold (unknown ).
struct ProbabilityGrid2BasicMap
{
  int occthreshold_;
  BasicMapCell defcell_; // This is what unknown will be converted to
  BasicMapCell occupied_;
  BasicMapCell free_;
public:

  ProbabilityGrid2BasicMap(int threshold = 50, BasicMapCell::t_occupancy defcell = BasicMapCell::occupied):
  occthreshold_(threshold), defcell_(BasicMapCell(defcell)),
  occupied_(BasicMapCell(BasicMapCell::occupied)), free_(BasicMapCell(BasicMapCell::free)){}

  inline BasicMapCell& operator()(const int& gridval)
  {
    if (gridval < 0){ return defcell_; }
    else if (gridval >= occthreshold_){ return occupied_; }
    return free_;
  }
};

typedef boost::multi_array<BasicMapCell, 2> t_basicmap;


#endif /* BASIC_MAP_H_ */
