/*
 * map.cpp
 *
 *  Created on: 17/08/2011
 *      Author: daver
 */

#include <boost/assert.hpp>
#include "map.h"

void occmap_copy_to(const t_occmap& input, t_occmap& output,
                    const IntPoint2D& new_origin, std::vector<IntPoint2D>& newpoints)
{
  const limits& lim = limits::instance();

  // Make sure the input + shift points fit in the output grid
  int out_xsize = output.shape()[0];
  int out_ysize = output.shape()[1];
  int in_new_maxx = new_origin.x() + input.shape()[0] - 1;
  int in_new_maxy = new_origin.y() + input.shape()[1] - 1;

  for (int out_x = 0; out_x < out_xsize; ++out_x)
    {
      for (int out_y = 0; out_y < out_ysize; ++out_y)
        {
          if ((out_x >= new_origin.x() && out_x <= in_new_maxx) && (out_y
              >= new_origin.y() && out_y <= in_new_maxy))
            {
              int in_x = out_x - new_origin.x();
              int in_y = out_y - new_origin.y();
              output[out_x][out_y] = input[in_x][in_y];
              if (output[out_x][out_y].ancestor_ != lim.c_invalid_point)
                output[out_x][out_y].ancestor_.shift(new_origin);
            }
          else
            {
              newpoints.push_back(IntPoint2D(out_x, out_y));
              output[out_x][out_y].occupancy_ = OccMapCell::unknown;
              output[out_x][out_y].unknown_ = OccMapCell::assume_occupied;
              output[out_x][out_y].dist_ = 0;
              output[out_x][out_y].ancestor_ = lim.c_invalid_point;
            }
        }
    }
}

OccMapUpdater::OccMapUpdater(int threshold):
    thresholdfn_(threshold), propagateOccupied_(true),
    propagateFree_(true), propagateUnknown_(true)
{

}


void OccMapUpdater::update(t_occmap& occmap, const t_occgrid& occgrid)
{
  std::vector<IntPoint2D> points;
  all_points(occgrid, points);
  update(occmap, occgrid, points.begin(), points.end());
}

void OccMapUpdater::clear(t_bushfire& bf, const int& val, const IntPoint2D& p)
{
  if (pu_visited_.find(p) != pu_visited_.end()) return;
  pu_visited_.insert(p);
  // Propagate to adjacent cells
  adjacents_.clear();
  adjacents4(*occmap_, p, adjacents_);
  BOOST_FOREACH(const IntPoint2D& ap, adjacents_)
  {
    if (pu_visited_.find(ap) != pu_visited_.end()) continue;

    OccMapCell& apbmc = get_cell(*occmap_, ap);
//    if (apbmc.occupancy_ == OccMapCell::unknown &&
//        (apbmc.ancestor_ == p || apbmc.ancestor_ == lim.c_invalid_point))
      if (apbmc.occupancy_ == OccMapCell::unknown && apbmc.ancestor_ == p)
    {
      // changes_.insert(ap);
      apbmc.setUnknown();
      propagateUnknown_.push(0, ap);
    }
    else if (apbmc.isOccupied())
    {
     propagateOccupied_.push(apbmc.dist_, ap);
    }
    else if (apbmc.isFree())
    {
     propagateFree_.push(apbmc.dist_, ap);
    }
  }
}

void OccMapUpdater::fillOccupied(t_bushfire& bf, const int& val, const IntPoint2D& p)
{
  OccMapCell& pbmc = get_cell(*occmap_, p);
  if (!pbmc.isOccupied()) return;
  int newval = val+1;

  // Propagate to adjacent cells
  adjacents_.clear();
  adjacents4(*occmap_, p, adjacents_);
  BOOST_FOREACH(const IntPoint2D& ap, adjacents_)
  {
    OccMapCell& apbmc = get_cell(*occmap_, ap);
    if (apbmc.occupancy_ != OccMapCell::unknown) continue;
    if (apbmc.unknown_ == OccMapCell::assume_occupied)
    {
      if (newval < apbmc.dist_)
      {
        apbmc.dist_ = newval;
        apbmc.ancestor_ = p;
        propagateOccupied_.push(newval, ap);
       }
      continue;
    }
    else
    {
      apbmc.unknown_ = OccMapCell::assume_occupied;
      apbmc.dist_ = newval;
      apbmc.ancestor_ = p;
      changes_.insert(ap);
      propagateOccupied_.push(newval, ap);
    }
  }
}

void OccMapUpdater::fillFree(t_bushfire& bf, const int& val, const IntPoint2D& p)
{
  OccMapCell& pbmc = get_cell(*occmap_, p);
  if (pbmc.isOccupied()) return; // The unknown cell may have been turned into an assumed_occupied by fillOccupied
  int newval = val+1;

  // Propagate to adjacent cells
  adjacents_.clear();
  adjacents4(*occmap_, p, adjacents_);
  BOOST_FOREACH(const IntPoint2D& ap, adjacents_)
  {
    OccMapCell& apbmc = get_cell(*occmap_, ap);
//    if (apbmc.isOccupied()) continue;
    if (apbmc.isOccupied()) return;
    if (apbmc.occupancy_ == OccMapCell::free) continue;
    if (apbmc.unknown_ == OccMapCell::assume_free)
    {
      if (newval < apbmc.dist_)
      {
        apbmc.dist_ = newval;
        apbmc.ancestor_ = p;
        propagateFree_.push(newval, ap);
       }
      continue;
    }
    apbmc.unknown_ = OccMapCell::assume_free;
    apbmc.dist_ = newval;
    apbmc.ancestor_ = p;
    changes_.insert(ap);
    propagateFree_.push(newval, ap);
  }
}

